@@ -2,6 +2,7 @@
 // SPDX-License-Identifier: Apache-2.0
 
 #include "common/common.h"
+#include "common/timer.h"
 #include "utils/image_buffer.h"
 #include "utils/random.h"
 #include <cassert>
@@ -83,12 +84,9 @@ TEST_CASE("physical device", "[physical_device]")
       else
         REQUIRE(getError() == Error::InvalidArgument);
 
-      const uint32_t nodeMask = physicalDevice.get<uint32_t>("nodeMask");
+      physicalDevice.get<uint32_t>("nodeMask");
       if (luidSupported)
-      {
         REQUIRE(getError() == Error::None);
-        REQUIRE(nodeMask != 0);
-      }
       else
         REQUIRE(getError() == Error::InvalidArgument);
 
@@ -124,19 +122,42 @@ TEST_CASE("physical device", "[physical_device]")
 
 // -------------------------------------------------------------------------------------------------
 
+TEST_CASE("device creation", "[device]")
+{
+  DeviceRef device = makeDevice();
+
+  SECTION("do nothing")
+  {
+  }
+
+  SECTION("get error")
+  {
+    REQUIRE(device.getError() == Error::None);
+  }
+
+  SECTION("commit")
+  {
+    device.commit();
+    REQUIRE(device.getError() == Error::None);
+  }
+}
+
+// -------------------------------------------------------------------------------------------------
+
 TEST_CASE("buffer creation", "[buffer]")
 {
+  const size_t bufferSize = 1024*768*3*4; // Metal requires shared buffer size to be multiple of 16K
   DeviceRef device = makeAndCommitDevice();
 
   SECTION("default buffer")
   {
-    BufferRef buffer = device.newBuffer(1234567);
+    BufferRef buffer = device.newBuffer(bufferSize);
     REQUIRE(device.getError() == Error::None);
   }
 
   SECTION("device buffer")
   {
-    BufferRef buffer = device.newBuffer(1234567, Storage::Device);
+    BufferRef buffer = device.newBuffer(bufferSize, Storage::Device);
     REQUIRE(device.getError() == Error::None);
   }
 
@@ -147,11 +168,20 @@ TEST_CASE("buffer creation", "[buffer]")
 
     if (managedMemorySupported)
     {
-      BufferRef buffer = device.newBuffer(1234567, Storage::Managed);
+      BufferRef buffer = device.newBuffer(bufferSize, Storage::Managed);
       REQUIRE(device.getError() == Error::None);
     }
   }
 
+  SECTION("shared buffer")
+  {
+    BufferRef buffer = device.newBuffer(bufferSize);
+    REQUIRE(device.getError() == Error::None);
+
+    BufferRef sharedBuffer = device.newBuffer(buffer.getData(), buffer.getSize());
+    REQUIRE(device.getError() == Error::None);
+  }
+
   SECTION("zero-sized default buffer")
   {
     BufferRef buffer = device.newBuffer(0);
@@ -164,6 +194,15 @@ TEST_CASE("buffer creation", "[buffer]")
     REQUIRE(device.getError() == Error::None);
   }
 
+  SECTION("zero-sized shared buffer")
+  {
+    BufferRef buffer = device.newBuffer(bufferSize);
+    REQUIRE(device.getError() == Error::None);
+
+    BufferRef sharedBuffer = device.newBuffer(buffer.getData(), 0);
+    REQUIRE(device.getError() == Error::None);
+  }
+
 #if !defined(OIDN_SANITIZER)
   SECTION("out-of-memory default buffer")
   {
@@ -180,13 +219,13 @@ TEST_CASE("buffer creation", "[buffer]")
 
   SECTION("invalid buffer storage")
   {
-    BufferRef buffer = device.newBuffer(10000, static_cast<Storage>(-42));
+    BufferRef buffer = device.newBuffer(bufferSize, static_cast<Storage>(-42));
     REQUIRE(device.getError() == Error::InvalidArgument);
   }
 
   SECTION("device released before buffer")
   {
-    BufferRef buffer = device.newBuffer(123456);
+    BufferRef buffer = device.newBuffer(bufferSize);
     REQUIRE(device.getError() == Error::None);
     device.release();
   }
@@ -306,6 +345,17 @@ std::shared_ptr<ImageBuffer> makeConstImage(DeviceRef& device, int W, int H, int
   return image;
 }
 
+std::shared_ptr<ImageBuffer> makeRandomImage(DeviceRef& device, int W, int H, int C = 3,
+                                             DataType dataType = DataType::Float32,
+                                             float minValue = 0.f, float maxValue = 1.f)
+{
+  Random rng;
+  auto image = std::make_shared<ImageBuffer>(device, W, H, C, dataType);
+  for (size_t i = 0; i < image->getSize(); ++i)
+    image->set(i, minValue + rng.getFloat() * (maxValue - minValue));
+  return image;
+}
+
 bool isBetween(const std::shared_ptr<ImageBuffer>& image, float a, float b)
 {
   for (size_t i = 0; i < image->getSize(); ++i)
@@ -529,7 +579,32 @@ TEST_CASE("multiple filters", "[multi_filter]")
 
 // -------------------------------------------------------------------------------------------------
 
-TEST_CASE("multiple devices", "[multi_device]")
+TEST_CASE("single device", "[single_device][minimal]")
+{
+  const std::vector<int> sizes = {111, 256, 80};
+
+  for (size_t i = 0; i < sizes.size(); ++i)
+  {
+    DeviceRef device = makeAndCommitDevice();
+
+    FilterRef filter = device.newFilter("RT");
+    REQUIRE(bool(filter));
+
+    auto image = makeConstImage(device, sizes[i], sizes[i], 3, DataType::Float16, 0.5f);
+    setFilterImage(filter, "color",  image);
+    setFilterImage(filter, "output", image);
+
+    filter.set("hdr", true);
+    filter.commit();
+    REQUIRE(device.getError() == Error::None);
+
+    filter.execute();
+    REQUIRE(device.getError() == Error::None);
+    REQUIRE(isBetween(image, 0.1f, 1.0f)); // output sanity check
+  }
+}
+
+TEST_CASE("multiple devices", "[multi_device][minimal]")
 {
   const std::vector<int> sizes = {111, 256, 80};
 
@@ -553,11 +628,28 @@ TEST_CASE("multiple devices", "[multi_device]")
     REQUIRE(devices[i].getError() == Error::None);
   }
 
-  for (size_t i = 0; i < devices.size(); ++i)
+  SECTION("lazy release")
   {
-    filters[i].execute();
-    REQUIRE(devices[i].getError() == Error::None);
-    REQUIRE(isBetween(images[i], 0.1f, 1.0f)); // output sanity check
+    for (size_t i = 0; i < devices.size(); ++i)
+    {
+      filters[i].execute();
+      REQUIRE(devices[i].getError() == Error::None);
+      REQUIRE(isBetween(images[i], 0.1f, 1.0f)); // output sanity check
+    }
+  }
+
+  SECTION("eager release")
+  {
+    for (size_t i = 0; i < devices.size(); ++i)
+    {
+      filters[i].execute();
+      REQUIRE(devices[i].getError() == Error::None);
+      REQUIRE(isBetween(images[i], 0.1f, 1.0f)); // output sanity check
+
+      images[i].reset();
+      filters[i].release();
+      devices[i].release();
+    }
   }
 }
 
@@ -622,6 +714,68 @@ TEST_CASE("shared image", "[shared_image]")
 
 // -------------------------------------------------------------------------------------------------
 
+TEST_CASE("inplace filter", "[inplace_filter]")
+{
+  // Metal requires shared buffer size to be multiple of 16K
+  const int W = 1920;
+  const int H = 1080;
+
+  DeviceRef device = makeAndCommitDevice();
+
+  FilterRef filter = device.newFilter("RT");
+  REQUIRE(bool(filter));
+
+  auto refColor  = makeRandomImage(device, W, H);
+  auto refOutput = makeImage(device, W, H);
+
+  setFilterImage(filter, "color",  refColor);
+  setFilterImage(filter, "output", refOutput);
+
+  filter.set("hdr", true);
+  filter.set("maxMemoryMB", 0); // make sure there will be multiple tiles
+
+  filter.commit();
+  REQUIRE(device.getError() == Error::None);
+
+  filter.execute();
+  REQUIRE(device.getError() == Error::None);
+
+  SECTION("same buffer")
+  {
+    auto color = refColor->clone();
+    setFilterImage(filter, "color", color);
+    setFilterImage(filter, "output", color);
+
+    filter.commit();
+    REQUIRE(device.getError() == Error::None);
+
+    filter.execute();
+    REQUIRE(device.getError() == Error::None);
+
+    REQUIRE(compareImage(*color, *refOutput));
+  }
+
+  SECTION("aliased shared buffers")
+  {
+    auto color = refColor->clone();
+    BufferRef colorBuffer  = color->getBuffer();
+    BufferRef outputBuffer = device.newBuffer(colorBuffer.getData(), colorBuffer.getSize());
+
+    filter.setImage("color",  colorBuffer,  color->getFormat(), W, H);
+    filter.setImage("output", outputBuffer, color->getFormat(), W, H);
+
+    filter.commit();
+    REQUIRE(device.getError() == Error::None);
+
+    filter.execute();
+    REQUIRE(device.getError() == Error::None);
+
+    REQUIRE(compareImage(*color, *refOutput));
+  }
+}
+
+// -------------------------------------------------------------------------------------------------
+
 TEST_CASE("filter update", "[filter_update]")
 {
   const int W = 211;
@@ -718,6 +872,7 @@ TEST_CASE("filter update", "[filter_update]")
 
 TEST_CASE("async filter", "[async_filter]")
 {
+  // Use a small image (one tile) to avoid potential blocking when filtering asynchronously on GPUs
   const int W = 799;
   const int H = 601;
 
@@ -737,9 +892,28 @@ TEST_CASE("async filter", "[async_filter]")
   filter.commit();
   REQUIRE(device.getError() == Error::None);
 
-  for (int i = 0; i < 3; ++i)
-    filter.executeAsync();
+  // Measure blocking filter time
+  double blockingTime = std::numeric_limits<double>::infinity();
+  for (int i = 0; i < 4; ++i)
+  {
+    Timer timer;
+    filter.execute();
+    blockingTime = min(blockingTime, timer.query());
+    REQUIRE(device.getError() == Error::None);
+  }
 
+  // Measure async filter time
+  double asyncTime = std::numeric_limits<double>::infinity();
+  for (int i = 0; i < 3; ++i)
+  {
+    Timer timer;
+    filter.executeAsync();
+    asyncTime = min(asyncTime, timer.query());
+    REQUIRE(device.getError() == Error::None);
+  }
+  REQUIRE(asyncTime < blockingTime / 2.); // async shouldn't block for long
+
+  // Change filter parameters without manually syncing first
   setFilterImage(filter, "albedo", albedo);
   filter.set("hdr", false);
 
@@ -753,6 +927,8 @@ TEST_CASE("async filter", "[async_filter]")
   REQUIRE(device.getError() == Error::None);
 
   filter.executeAsync();
+
+  // Release filter without manually syncing first
   filter.release();
 
   REQUIRE(device.getError() == Error::None);
@@ -863,25 +1039,65 @@ struct Progress
 {
   double n;    // current progress
   double nMax; // when to cancel execution
+  std::atomic<bool> running; // should the callback be called?
 
-  Progress(double nMax) : n(0), nMax(nMax) {}
+  Progress(double nMax)
+    : n(0),
+      nMax(nMax),
+      running(false) {}
 };
 
 // Progress monitor callback function
 bool progressCallback(void* userPtr, double n)
 {
   Progress* progress = (Progress*)userPtr;
+  REQUIRE(progress->running); // callback should be called only while running
   REQUIRE((std::isfinite(n) && n >= 0 && n <= 1)); // n must be between 0 and 1
   REQUIRE(n >= progress->n); // n must not decrease
   progress->n = n;
   return n < progress->nMax; // cancel if reached nMax
 }
 
-void progressTest(DeviceRef& device, double nMax = 1000)
+double progressTest(DeviceRef& device, FilterRef& filter, double nMax = 1000)
+{
+  Progress progress(nMax);
+  filter.setProgressMonitorFunction(progressCallback, &progress);
+
+  filter.commit();
+  REQUIRE(device.getError() == Error::None);
+
+  Timer timer;
+  progress.running = true;
+  filter.execute();
+  progress.running = false;
+  const double time = timer.query();
+
+  if (nMax <= 1)
+  {
+    // Execution should be cancelled but it's not guaranteed
+    Error error = device.getError();
+    REQUIRE((error == Error::None || error == Error::Cancelled));
+    REQUIRE((progress.n >= nMax && progress.n <= 1));
+  }
+  else
+  {
+    // Execution should be finished
+    REQUIRE(device.getError() == Error::None);
+    REQUIRE(progress.n == 1); // progress must be 100% at the end
+  }
+
+  return time;
+}
+
+TEST_CASE("progress monitor", "[progress]")
 {
   const int W = 1283;
   const int H = 727;
 
+  DeviceRef device = makeAndCommitDevice();
+  const bool isCPU = device.get<DeviceType>("type") == DeviceType::CPU;
+  REQUIRE(device.getError() == Error::None);
+
   FilterRef filter = device.newFilter("RT");
   REQUIRE(bool(filter));
 
@@ -889,54 +1105,62 @@ void progressTest(DeviceRef& device, double nMax = 1000)
   setFilterImage(filter, "color",  image);
   setFilterImage(filter, "output", image); // in-place
 
-  Progress progress(nMax);
-  filter.setProgressMonitorFunction(progressCallback, &progress);
-
   filter.set("maxMemoryMB", 0); // make sure there will be multiple tiles
 
   filter.commit();
   REQUIRE(device.getError() == Error::None);
 
-  filter.execute();
-
-  if (nMax <= 1)
+  // Measure filter time without progress monitor
+  double refTime = std::numeric_limits<double>::infinity();
+  for (int i = 0; i < 4; ++i)
   {
-    // Execution should be cancelled but it's not guaranteed
-    Error error = device.getError();
-    REQUIRE((error == Error::None || error == Error::Cancelled));
-    // Check whether the callback has not been called after requesting cancellation
-    REQUIRE(progress.n >= nMax);
-  }
-  else
-  {
-    // Execution should be finished
+    Timer timer;
+    filter.execute();
+    refTime = min(refTime, timer.query());
     REQUIRE(device.getError() == Error::None);
-    REQUIRE(progress.n == 1); // progress must be 100% at the end
   }
-}
 
-TEST_CASE("progress monitor", "[progress]")
-{
-  DeviceRef device = makeAndCommitDevice();
-
-  SECTION("progress monitor: complete")
+  SECTION("progress monitor: finish")
   {
-    progressTest(device);
+    progressTest(device, filter);
   }
 
   SECTION("progress monitor: cancel at the middle")
   {
-    progressTest(device, 0.5);
+    progressTest(device, filter, 0.5);
+  }
+
+  SECTION("progress monitor: cancel early")
+  {
+    double time = std::numeric_limits<double>::infinity();
+    for (int i = 0; i < 3; ++i)
+      time = min(time, progressTest(device, filter, 0.1));
+    REQUIRE((!isCPU || time < refTime / 2.)); // CPU should be able to cancel quickly
   }
 
   SECTION("progress monitor: cancel at the beginning")
   {
-    progressTest(device, 0);
+    double time = std::numeric_limits<double>::infinity();
+    for (int i = 0; i < 3; ++i)
+      time = min(time, progressTest(device, filter, 0));
+    REQUIRE((!isCPU || time < refTime / 2.)); // CPU should be able to cancel quickly
   }
 
   SECTION("progress monitor: cancel at the end")
   {
-    progressTest(device, 1);
+    progressTest(device, filter, 1);
+  }
+
+  SECTION("progress monitor: cancel around the middle, finish")
+  {
+    progressTest(device, filter, 0.4);
+    progressTest(device, filter);
+  }
+
+  SECTION("progress monitor: finish, cancel around the middle")
+  {
+    progressTest(device, filter);
+    progressTest(device, filter, 0.6);
   }
 }
 
