@@ -8,11 +8,7 @@ OIDN_NAMESPACE_BEGIN
 
   UNetFilter::UNetFilter(const Ref<Device>& device)
     : Filter(device)
-  {
-    // Compute final device-dependent tile alignment and overlap
-    tileAlignment = lcm(minTileAlignment, device->getMinTileAlignment());
-    tileOverlap = round_up(receptiveField / 2, tileAlignment);
-  }
+  {}
 
   void UNetFilter::setData(const std::string& name, const Data& data)
   {
@@ -51,7 +47,8 @@ OIDN_NAMESPACE_BEGIN
       Quality qualityValue = static_cast<Quality>(value);
       if (qualityValue == Quality::Default)
         qualityValue = defaultQuality;
-      else if (qualityValue != Quality::High && qualityValue != Quality::Balanced)
+      else if (qualityValue != Quality::High && qualityValue != Quality::Balanced &&
+               qualityValue != Quality::Fast)
         throw Exception(Error::InvalidArgument, "unknown filter quality mode");
       setParam(quality, qualityValue);
     }
@@ -130,11 +127,10 @@ OIDN_NAMESPACE_BEGIN
     if (dirtyParam)
     {
       // Make sure that all asynchronous operations have completed
-      device->wait();
+      device->waitAndThrow();
 
       // (Re-)Initialize the filter
-      device->getEngine()->runHostTask([&]() { init(); });
-      device->wait();
+      device->execute([&]() { init(); });
 
       // Clean up the device memory if the memory usage limit has been reduced
       if (maxMemoryMB >= 0 && (maxMemoryMB < prevMaxMemoryMB || prevMaxMemoryMB < 0))
@@ -154,17 +150,23 @@ OIDN_NAMESPACE_BEGIN
     if (H <= 0 || W <= 0)
       return;
 
-    auto mainEngine = device->getEngine();
-
-    mainEngine->runHostTask([&]()
+    device->execute([&]()
     {
       // Initialize the progress state
-      double workAmount = tileCountH * tileCountW * instances[0].graph->getWorkAmount();
-      if (hdr && math::isnan(inputScale))
-        workAmount += 1;
-      if (outputTemp)
-        workAmount += 1;
-      progress.start(mainEngine, progressFunc, progressUserPtr, workAmount);
+      Ref<Progress> progress;
+      if (progressFunc)
+      {
+        size_t workAmount = 0;
+        for (int i = 0; i < device->getNumSubdevices(); ++i)
+          workAmount += instances[i].graph->getWorkAmount();
+        workAmount *= (tileCountH * tileCountW) / device->getNumSubdevices();
+        if (hdr && math::isnan(inputScale))
+          workAmount += autoexposure->getWorkAmount();
+        if (outputTemp)
+          workAmount += imageCopy->getWorkAmount();
+
+        progress = makeRef<Progress>(progressFunc, progressUserPtr, workAmount);
+      }
 
       // Set the input scale
       if (math::isnan(inputScale))
@@ -172,9 +174,8 @@ OIDN_NAMESPACE_BEGIN
         if (hdr)
         {
           autoexposure->setSrc(color);
-          autoexposure->submit();
+          autoexposure->submit(progress);
           device->submitBarrier();
-          progress.update(mainEngine, 1);
           transferFunc->setInputScale(autoexposure->getDstPtr());
         }
         else
@@ -232,7 +233,7 @@ OIDN_NAMESPACE_BEGIN
           //printf("Tile: %d %d -> %d %d\n", w+overlapBeginW, h+overlapBeginH, w+overlapBeginW+tileW2, h+overlapBeginH+tileH2);
 
           // Denoise the tile
-          instance.graph->run(progress);
+          instance.graph->submit(progress);
 
           // Next tile
           tileIndex++;
@@ -245,17 +246,9 @@ OIDN_NAMESPACE_BEGIN
       if (outputTemp)
       {
         imageCopy->setDst(output);
-        imageCopy->submit();
+        imageCopy->submit(progress);
       }
-
-      // Finished
-      progress.finish(mainEngine);
-    });
-
-    if (sync == SyncMode::Sync)
-      device->wait();
-    else
-      device->flush();
+    }, sync);
   }
 
   void UNetFilter::init()
@@ -263,11 +256,18 @@ OIDN_NAMESPACE_BEGIN
     cleanup();
     checkParams();
 
-    // Build the model
+    // Select the model
     Data weightsBlob = getWeights();
     auto constTensors = parseTZA(weightsBlob.ptr, weightsBlob.size);
-    const bool fastMath = quality == Quality::Balanced;
+    const bool fastMath = quality != Quality::High;
+    largeModel = constTensors->find("enc_conv1b.weight") != constTensors->end();
 
+    // Compute final device-dependent tile alignment and overlap
+    const int receptiveField = largeModel ? receptiveFieldLarge : receptiveFieldBase;
+    tileAlignment = lcm(minTileAlignment, device->getMinTileAlignment());
+    tileOverlap = round_up(receptiveField / 2, tileAlignment);
+
+    // Build the model
     for (int i = 0; i < device->getNumSubdevices(); ++i)
     {
       Engine* engine = device->getEngine(i);
@@ -275,7 +275,7 @@ OIDN_NAMESPACE_BEGIN
       // We can use cached weights only for built-in weights because user weights may change!
       auto cachedConstTensors =
         userWeightsBlob ? nullptr : engine->getSubdevice()->getCachedTensors(weightsBlob.ptr);
-      
+
       instances.emplace_back();
       instances.back().graph = makeRef<Graph>(engine, constTensors, cachedConstTensors, fastMath);
     }
@@ -393,25 +393,25 @@ OIDN_NAMESPACE_BEGIN
 
   Data UNetFilter::getWeights()
   {
-    // Select the weights to use
-    Data weightsBlob;
+    // Select the model to use
+    Model* model = nullptr;
 
     if (color)
     {
       if (!albedo && !normal)
       {
-        weightsBlob = directional ? weightsBlobs.dir : (hdr ? weightsBlobs.hdr : weightsBlobs.ldr);
+        model = directional ? &models.dir : (hdr ? &models.hdr : &models.ldr);
       }
       else if (albedo && !normal)
       {
-        weightsBlob = hdr ? weightsBlobs.hdr_alb : weightsBlobs.ldr_alb;
+        model = hdr ? &models.hdr_alb : &models.ldr_alb;
       }
       else if (albedo && normal)
       {
         if (cleanAux)
-          weightsBlob = hdr ? weightsBlobs.hdr_calb_cnrm : weightsBlobs.ldr_calb_cnrm;
+          model = hdr ? &models.hdr_calb_cnrm : &models.ldr_calb_cnrm;
         else
-          weightsBlob = hdr ? weightsBlobs.hdr_alb_nrm : weightsBlobs.ldr_alb_nrm;
+          model = hdr ? &models.hdr_alb_nrm : &models.ldr_alb_nrm;
       }
     }
     else
@@ -421,13 +421,13 @@ OIDN_NAMESPACE_BEGIN
       {
         if (hdr)
           throw Exception(Error::InvalidOperation, "hdr mode is not supported for albedo filtering");
-        weightsBlob = weightsBlobs.alb;
+        model = &models.alb;
       }
       else if (!albedo && normal)
       {
         if (hdr || srgb)
           throw Exception(Error::InvalidOperation, "hdr and srgb modes are not supported for normal filtering");
-        weightsBlob = weightsBlobs.nrm;
+        model = &models.nrm;
       }
       else
       {
@@ -435,8 +435,29 @@ OIDN_NAMESPACE_BEGIN
       }
     }
 
+    // Select the weights to use
+    Data weightsBlob = nullptr;
+
     if (userWeightsBlob)
+    {
       weightsBlob = userWeightsBlob;
+    }
+    else if (model)
+    {
+      switch (quality)
+      {
+      case Quality::Default:
+      case Quality::High:
+        weightsBlob = model->large ? model->large : model->base;
+        break;
+      case Quality::Balanced:
+        weightsBlob = model->base;
+        break;
+      case Quality::Fast:
+        weightsBlob = model->small ? model->small : model->base;
+        break;
+      }
+    }
 
     if (!weightsBlob)
       throw Exception(Error::InvalidOperation, "unsupported combination of input features");
@@ -444,6 +465,71 @@ OIDN_NAMESPACE_BEGIN
     return weightsBlob;
   }
 
+  Ref<Op> UNetFilter::addUNet(const Ref<Graph>& graph, const Ref<Op>& inputProcess)
+  {
+    auto x = graph->addConv("enc_conv0", inputProcess, Activation::ReLU);
+
+    auto pool1 = x = graph->addConv("enc_conv1", x, Activation::ReLU, PostOp::Pool);
+
+    auto pool2 = x = graph->addConv("enc_conv2", x, Activation::ReLU, PostOp::Pool);
+
+    auto pool3 = x = graph->addConv("enc_conv3", x, Activation::ReLU, PostOp::Pool);
+
+    auto pool4 = x = graph->addConv("enc_conv4", x, Activation::ReLU, PostOp::Pool);
+
+    x = graph->addConv("enc_conv5a", pool4, Activation::ReLU);
+    x = graph->addConv("enc_conv5b", x, Activation::ReLU, PostOp::Upsample);
+
+    x = graph->addConcatConv("dec_conv4a", x, pool3, Activation::ReLU);
+    x = graph->addConv("dec_conv4b", x, Activation::ReLU, PostOp::Upsample);
+
+    x = graph->addConcatConv("dec_conv3a", x, pool2, Activation::ReLU);
+    x = graph->addConv("dec_conv3b", x, Activation::ReLU, PostOp::Upsample);
+
+    x = graph->addConcatConv("dec_conv2a", x, pool1, Activation::ReLU);
+    x = graph->addConv("dec_conv2b", x, Activation::ReLU, PostOp::Upsample);
+
+    x = graph->addConcatConv("dec_conv1a", x, inputProcess, Activation::ReLU);
+    x = graph->addConv("dec_conv1b", x, Activation::ReLU);
+
+    x = graph->addConv("dec_conv0", x, Activation::ReLU);
+
+    return x;
+  }
+
+  Ref<Op> UNetFilter::addUNetLarge(const Ref<Graph>& graph, const Ref<Op>& inputProcess)
+  {
+    auto x = graph->addConv("enc_conv1a", inputProcess, Activation::ReLU);
+    auto pool1 = x = graph->addConv("enc_conv1b", x, Activation::ReLU, PostOp::Pool);
+
+    x = graph->addConv("enc_conv2a", x, Activation::ReLU);
+    auto pool2 = x = graph->addConv("enc_conv2b", x, Activation::ReLU, PostOp::Pool);
+
+    x = graph->addConv("enc_conv3a", x, Activation::ReLU);
+    auto pool3 = x = graph->addConv("enc_conv3b", x, Activation::ReLU, PostOp::Pool);
+
+    x = graph->addConv("enc_conv4a", x, Activation::ReLU);
+    auto pool4 = x = graph->addConv("enc_conv4b", x, Activation::ReLU, PostOp::Pool);
+
+    x = graph->addConv("enc_conv5a", pool4, Activation::ReLU);
+    x = graph->addConv("enc_conv5b", x, Activation::ReLU, PostOp::Upsample);
+
+    x = graph->addConcatConv("dec_conv4a", x, pool3, Activation::ReLU);
+    x = graph->addConv("dec_conv4b", x, Activation::ReLU, PostOp::Upsample);
+
+    x = graph->addConcatConv("dec_conv3a", x, pool2, Activation::ReLU);
+    x = graph->addConv("dec_conv3b", x, Activation::ReLU, PostOp::Upsample);
+
+    x = graph->addConcatConv("dec_conv2a", x, pool1, Activation::ReLU);
+    x = graph->addConv("dec_conv2b", x, Activation::ReLU, PostOp::Upsample);
+
+    x = graph->addConcatConv("dec_conv1a", x, inputProcess, Activation::ReLU);
+    x = graph->addConv("dec_conv1b", x, Activation::ReLU);
+    x = graph->addConv("dec_conv1c", x, Activation::ReLU);
+
+    return x;
+  }
+
   // Tries to build the model without exceeding the specified amount of memory
   bool UNetFilter::buildModel(size_t maxMemoryByteSize)
   {
@@ -473,37 +559,9 @@ OIDN_NAMESPACE_BEGIN
       auto& graph = instance.graph;
 
       // Create the model graph
-      auto inputProcess = graph->addInputProcess("input", inputDims,
-                                                 transferFunc, hdr, snorm);
-
-      auto encConv0 = graph->addConv("enc_conv0", inputProcess, Activation::ReLU);
-
-      auto pool1 = graph->addConv("enc_conv1", encConv0, Activation::ReLU, PostOp::Pool);
-
-      auto pool2 = graph->addConv("enc_conv2", pool1, Activation::ReLU, PostOp::Pool);
-
-      auto pool3 = graph->addConv("enc_conv3", pool2, Activation::ReLU, PostOp::Pool);
-
-      auto pool4 = graph->addConv("enc_conv4", pool3, Activation::ReLU, PostOp::Pool);
-
-      auto encConv5a = graph->addConv("enc_conv5a", pool4, Activation::ReLU);
-
-      auto upsample4 = graph->addConv("enc_conv5b", encConv5a, Activation::ReLU, PostOp::Upsample);
-      auto decConv4a = graph->addConcatConv("dec_conv4a", upsample4, pool3, Activation::ReLU);
-
-      auto upsample3 = graph->addConv("dec_conv4b", decConv4a, Activation::ReLU, PostOp::Upsample);
-      auto decConv3a = graph->addConcatConv("dec_conv3a", upsample3, pool2, Activation::ReLU);
-
-      auto upsample2 = graph->addConv("dec_conv3b", decConv3a, Activation::ReLU, PostOp::Upsample);
-      auto decConv2a = graph->addConcatConv("dec_conv2a", upsample2, pool1, Activation::ReLU);
-
-      auto upsample1 = graph->addConv("dec_conv2b", decConv2a, Activation::ReLU, PostOp::Upsample);
-      auto decConv1a = graph->addConcatConv("dec_conv1a", upsample1, inputProcess, Activation::ReLU);
-      auto decConv1b = graph->addConv("dec_conv1b", decConv1a, Activation::ReLU);
-
-      auto decConv0 = graph->addConv("dec_conv0", decConv1b, Activation::ReLU);
-
-      auto outputProcess = graph->addOutputProcess("output", decConv0, transferFunc, hdr, snorm);
+      auto inputProcess = graph->addInputProcess("input", inputDims, transferFunc, hdr, snorm);
+      auto x = largeModel ? addUNetLarge(graph, inputProcess) : addUNet(graph, inputProcess);
+      auto outputProcess = graph->addOutputProcess("output", x, transferFunc, hdr, snorm);
 
       // Check whether all operations in the graph are supported
       if (!graph->isSupported())
