@@ -12,7 +12,12 @@ OIDN_NAMESPACE_BEGIN
   {
   public:
     CPUAutoexposure(CPUEngine* engine, const ImageDesc& srcDesc);
-    void submit() override;
+
+    Engine* getEngine() const override { return engine; }
+    void submitKernels(const Ref<CancellationToken>& ct) override;
+
+  private:
+    CPUEngine* engine;
   };
 
 OIDN_NAMESPACE_END
