@@ -12,9 +12,12 @@ OIDN_NAMESPACE_BEGIN
   {
   public:
     CPUConv(CPUEngine* engine, const ConvDesc& desc);
-    void submit() override;
+
+    Engine* getEngine() const override { return engine; }
+    void submitKernels(const Ref<CancellationToken>& ct) override;
 
   private:
+    CPUEngine* engine;
     int blockOCB; // block of output channel blocks
     int blockOW;  // block of output width
     int OCBB;     // number of output channel block blocks
