@@ -8,9 +8,10 @@
 OIDN_NAMESPACE_BEGIN
 
   CPUImageCopy::CPUImageCopy(CPUEngine* engine)
+    : engine(engine)
   {}
 
-  void CPUImageCopy::submit()
+  void CPUImageCopy::submitKernels(const Ref<CancellationToken>& ct)
   {
     check();
 
@@ -18,10 +19,13 @@ OIDN_NAMESPACE_BEGIN
     kernel.src = *src;
     kernel.dst = *dst;
 
-    parallel_nd(dst->getH(), [&](int h)
+    engine->submitFunc([=]
     {
-      ispc::CPUImageCopyKernel_run(&kernel, h);
-    });
+      parallel_for(kernel.dst.H, [&](int h)
+      {
+        ispc::CPUImageCopyKernel_run(&kernel, h);
+      });
+    }, ct);
   }
 
 OIDN_NAMESPACE_END
\ No newline at end of file
