@@ -8,9 +8,11 @@
 OIDN_NAMESPACE_BEGIN
 
   CPUInputProcess::CPUInputProcess(CPUEngine* engine, const InputProcessDesc& desc)
-    : InputProcess(engine, desc) {}
+    : InputProcess(engine, desc),
+      engine(engine)
+  {}
 
-  void CPUInputProcess::submit()
+  void CPUInputProcess::submitKernels(const Ref<CancellationToken>& ct)
   {
     check();
 
@@ -26,10 +28,13 @@ OIDN_NAMESPACE_BEGIN
     kernel.hdr   = hdr;
     kernel.snorm = snorm;
 
-    parallel_nd(kernel.dst.H, [&](int hDst)
+    engine->submitFunc([=]
     {
-      ispc::CPUInputProcessKernel_run(&kernel, hDst);
-    });
+      parallel_for(kernel.dst.H, [&](int hDst)
+      {
+        ispc::CPUInputProcessKernel_run(&kernel, hDst);
+      });
+    }, ct);
   }
 
 OIDN_NAMESPACE_END
\ No newline at end of file
