@@ -8,9 +8,11 @@
 OIDN_NAMESPACE_BEGIN
 
   CPUOutputProcess::CPUOutputProcess(CPUEngine* engine, const OutputProcessDesc& desc)
-    : OutputProcess(desc) {}
+    : OutputProcess(desc),
+      engine(engine)
+  {}
 
-  void CPUOutputProcess::submit()
+  void CPUOutputProcess::submitKernels(const Ref<CancellationToken>& ct)
   {
     check();
 
@@ -23,10 +25,13 @@ OIDN_NAMESPACE_BEGIN
     kernel.hdr = hdr;
     kernel.snorm = snorm;
 
-    parallel_nd(kernel.tile.H, [&](int h)
+    engine->submitFunc([=]
     {
-      ispc::CPUOutputProcessKernel_run(&kernel, h);
-    });
+      parallel_for(kernel.tile.H, [&](int h)
+      {
+        ispc::CPUOutputProcessKernel_run(&kernel, h);
+      });
+    }, ct);
   }
 
 OIDN_NAMESPACE_END
\ No newline at end of file
