@@ -8,14 +8,15 @@
 OIDN_NAMESPACE_BEGIN
 
   CPUPool::CPUPool(CPUEngine* engine, const PoolDesc& desc)
-    : Pool(desc)
+    : Pool(desc),
+      engine(engine)
   {
     if (srcDesc.layout != TensorLayout::Chw8c &&
         srcDesc.layout != TensorLayout::Chw16c)
       throw std::invalid_argument("unsupported pooling source layout");
   }
 
-  void CPUPool::submit()
+  void CPUPool::submitKernels(const Ref<CancellationToken>& ct)
   {
     if (!src || !dst)
       throw std::logic_error("pooling source/destination not set");
@@ -26,10 +27,13 @@ OIDN_NAMESPACE_BEGIN
     kernel.src = *src;
     kernel.dst = *dst;
 
-    parallel_nd(dst->getPaddedC() / blockC, dst->getH(), [&](int cb, int h)
+    engine->submitFunc([=]
     {
-      ispc::CPUPoolKernel_run(&kernel, cb, h);
-    });
+      parallel_for(kernel.dst.C / blockC, kernel.dst.H, [&](int cb, int h)
+      {
+        ispc::CPUPoolKernel_run(&kernel, cb, h);
+      });
+    }, ct);
   }
 
 OIDN_NAMESPACE_END
\ No newline at end of file
